@@ -58,6 +58,15 @@ typedef struct _FcitxICDataInfo {
     void* arg;
 } FcitxICDataInfo;
 
+typedef enum _FcitxEventFlag {
+    FEF_NONE = 0,
+    FEF_UI_MOVE = (1 << 1),
+    FEF_UI_UPDATE = (1 << 2),
+    FEF_EVENT_CHECK = (1 << 3),
+    FEF_RELOAD_ADDON = (1 << 4),
+    FEF_PROCESS_EVENT_MASK = ( FEF_UI_MOVE | FEF_UI_UPDATE | FEF_EVENT_CHECK )
+} FcitxUIFlag;
+
 struct _FcitxInstance {
     pthread_mutex_t fcitxMutex;
     UT_array uistats;
@@ -85,6 +94,8 @@ struct _FcitxInstance {
     struct _FcitxInputContext *ic_list;
     struct _FcitxInputContext *free_list;
     sem_t* sem;
+    sem_t startUpSem;
+    sem_t notifySem;
     pthread_t pid;
     fd_set rfds, wfds, efds;
     int maxfd;
@@ -92,6 +103,8 @@ struct _FcitxInstance {
 
     struct _HookStack* hookPreInputFilter;
     struct _HookStack* hookPostInputFilter;
+    struct _HookStack* hookPreReleaseInputFilter;
+    struct _HookStack* hookPostReleaseInputFilter;
     struct _HookStack* hookOutputFilter;
     struct _HookStack* hookHotkeyFilter;
     struct _HookStack* hookResetInputHook;
@@ -105,7 +118,7 @@ struct _FcitxInstance {
     struct _HookStack* hookICStateChangedHook;
     struct _HookStack* hookIMChangedHook;
 
-    FcitxUIFlag uiflag;
+    uint32_t eventflag;
 
     FcitxContextState globalState;
 
@@ -138,14 +151,17 @@ struct _FcitxInstance {
     UT_array* enableList;
 
     UT_array timeout;
-    boolean initialized;
+    volatile boolean initialized;
     uint64_t timeoutIdx;
 
     UT_array icdata;
-    boolean loadingFatalError;
-    boolean quietQuit;
-    boolean destroy;
+    volatile boolean loadingFatalError;
+    volatile boolean quietQuit;
+    volatile boolean destroy;
     int fd;
+    int overrideDelay;
+    
+    UT_array eventQueue;
 };
 
 static inline FcitxAddon**
