@@ -68,6 +68,28 @@ static Bool FcitxXkbUpdateProperties(FcitxXkb* xkb, const char *rules_file,
                                      const char *all_options);
 static void FcitxXkbApplyCustomScript(FcitxXkb* xkb);
 
+
+static inline boolean IMIsKeyboardIM(const char* imname) {
+    return strncmp(imname, "fcitx-keyboard-",
+                   strlen("fcitx-keyboard-")) == 0;
+}
+
+static void ExtractKeyboardIMLayout(const char* imname, char** layout, char** variant) {
+    if (!IMIsKeyboardIM(imname)) {
+        return;
+    }
+
+    imname += strlen("fcitx-keyboard-");
+    char* v = strchr(imname, '-');
+    if (v) {
+        *layout = strndup(imname, v - imname);
+        v++;
+        *variant = strdup(v);
+    } else {
+        *layout = strdup(imname);
+    }
+}
+
 #if 0
 static char* FcitxXkbGetCurrentLayout(FcitxXkb* xkb);
 static char* FcitxXkbGetCurrentModel(FcitxXkb* xkb);
@@ -75,7 +97,7 @@ static char* FcitxXkbGetCurrentOption(FcitxXkb* xkb);
 #endif
 static boolean
 FcitxXkbSetLayoutByName(FcitxXkb *xkb,
-                        const char *layout, const char *variant);
+                        const char *layout, const char *variant, boolean toDefault);
 static void
 FcitxXkbRetrieveCloseGroup(FcitxXkb *xkb);
 static boolean FcitxXkbSetLayout  (FcitxXkb* xkb,
@@ -176,37 +198,50 @@ static char* FcitxXkbFindXkbRulesFile(FcitxXkb* xkb)
         } else {
             int count = 0, i = 0;
             const char* base = XLIBDIR;
+            char *parent_to_free = NULL;
             while (base[i]) {
                 if (base[i] == '/')
                     count++;
                 i++;
             }
 
+            /**
+             * guess X11 data base directory.
+             **/
             if (count >= 3) {
                 // .../usr/lib/X11 -> /usr/share/X11/xkb vs
                 // .../usr/X11/lib -> /usr/X11/share/X11/xkb
                 const char* delta = StringEndsWith(base, "X11") ?
                     "/../../share/X11" : "/../share/X11";
-                char *tmppath;
-                fcitx_utils_alloc_cat_str(tmppath, base, delta,
-                                          "/xkb/rules/", rulesName, ".xml");
-                if(fcitx_utils_isreg(tmppath)) {
-                    rulesFile = realpath(tmppath, NULL);
-                    free(tmppath);
-                } else {
-                    fcitx_utils_alloc_cat_str(tmppath, base, "/X11/xkb/rules/",
-                                              rulesName, ".xml");
-                    if(fcitx_utils_isreg(tmppath)) {
-                        rulesFile = realpath(tmppath, NULL);
-                        free(tmppath);
+                fcitx_utils_alloc_cat_str(parent_to_free, base, delta);
+                if(!fcitx_utils_isdir(parent_to_free)) {
+                    // fallback to ${base}/X11
+                    fcitx_utils_set_cat_str(parent_to_free, base, "/X11");
+                    if(!fcitx_utils_isdir(parent_to_free)) {
+                        free(parent_to_free);
+                        parent_to_free = NULL;
                     }
                 }
             }
-            if(!rulesFile) {
-                fcitx_utils_alloc_cat_str(rulesFile,
-                                          "/usr/share/X11/xkb/rules/",
-                                          rulesName, ".xml");
+            const char *parent_path;
+            if (parent_to_free) {
+                /**
+                 * Found a existing dir, simplify it.
+                 * Using realpath() on rules files' name can change the base
+                 * name of the file (due to symlink), so it is only safe
+                 * to do it for directory's name. T-T..
+                 **/
+                char *tmp = realpath(parent_to_free, NULL);
+                parent_path = tmp;
+                free(parent_to_free);
+                parent_to_free = tmp;
+            } else {
+                // last fallback for known rules name.
+                parent_path = "/usr/share/X11";
             }
+            fcitx_utils_alloc_cat_str(rulesFile, parent_path,
+                                      "/xkb/rules/", rulesName, ".xml");
+            fcitx_utils_free(parent_to_free);
         }
         free(rulesName);
     } else {
@@ -219,7 +254,7 @@ static void
 FcitxXkbInitDefaultLayout(FcitxXkb* xkb)
 {
     Display* dpy = xkb->dpy;
-        XkbRF_VarDefsRec vd;
+    XkbRF_VarDefsRec vd;
 
     utarray_clear(xkb->defaultLayouts);
     utarray_clear(xkb->defaultModels);
@@ -291,14 +326,22 @@ FcitxXkbSetRules(FcitxXkb* xkb, const char *rules_file, const char *model,
     memset(&rnames, 0, sizeof(XkbComponentNamesRec));
     rdefs.model = model ? strdup(model) : NULL;
     rdefs.layout = all_layouts ? strdup(all_layouts) : NULL;
-    rdefs.variant = all_variants ? strdup(all_variants) : NULL;
-    rdefs.options = all_options ? strdup(all_options) : NULL;
+    rdefs.variant = all_variants && all_variants[0] ? strdup(all_variants) : NULL;
+    rdefs.options = all_options && all_options[0] ? strdup(all_options) : NULL;
     XkbRF_GetComponents(rules, &rdefs, &rnames);
     xkbDesc = XkbGetKeyboardByName(dpy, XkbUseCoreKbd, &rnames,
                                    XkbGBN_AllComponentsMask,
                                    XkbGBN_AllComponentsMask &
                                    (~XkbGBN_GeometryMask), True);
 
+    XkbRF_Free(rules, True);
+    free(rnames.keymap);
+    free(rnames.keycodes);
+    free(rnames.types);
+    free(rnames.compat);
+    free(rnames.symbols);
+    free(rnames.geometry);
+
     Bool result = True;
     if (!xkbDesc) {
         FcitxLog (WARNING, "Cannot load new keyboard description.");
@@ -308,6 +351,7 @@ FcitxXkbSetRules(FcitxXkb* xkb, const char *rules_file, const char *model,
         char* tempstr = strdup(rules_file);
         XkbRF_SetNamesProp(dpy, tempstr, &rdefs);
         free (tempstr);
+        XkbFreeKeyboard(xkbDesc, XkbGBN_AllComponentsMask, True);
     }
     free(rdefs.model);
     free(rdefs.layout);
@@ -373,16 +417,14 @@ FcitxXkbUpdateProperties(FcitxXkb* xkb, const char *rules_file,
         next += strlen (all_options);
     }
     *next++ = '\0';
-    if ((next - pval) != len) {
-        free (pval);
-        return True;
+    if ((next - pval) == len) {
+        XChangeProperty (dpy, root_window,
+                        rules_atom, XA_STRING, 8, PropModeReplace,
+                        (unsigned char *) pval, len);
+        XSync(dpy, False);
     }
 
-    XChangeProperty (dpy, root_window,
-                     rules_atom, XA_STRING, 8, PropModeReplace,
-                     (unsigned char *) pval, len);
-    XSync(dpy, False);
-
+    free(pval);
     return True;
 }
 
@@ -496,7 +538,7 @@ FcitxXkbGetCurrentGroup(FcitxXkb* xkb)
 }
 
 static void FcitxXkbAddNewLayout(FcitxXkb* xkb, const char* layoutString,
-                                 const char* variantString)
+                                 const char* variantString, boolean toDefault, int index)
 {
     if (!layoutString)
         return;
@@ -506,41 +548,64 @@ static void FcitxXkbAddNewLayout(FcitxXkb* xkb, const char* layoutString,
         const char* dummy = "";
         utarray_push_back(xkb->defaultVariants, &dummy);
     }
-    while (utarray_len(xkb->defaultVariants) >= 4) {
+
+    while (utarray_len(xkb->defaultVariants) >
+           utarray_len(xkb->defaultLayouts)) {
         utarray_pop_back(xkb->defaultVariants);
-        utarray_pop_back(xkb->defaultLayouts);
     }
-    utarray_push_back(xkb->defaultLayouts, &layoutString);
-    if (variantString)
-        utarray_push_back(xkb->defaultVariants, &variantString);
-    else {
-        const char* dummy = "";
-        utarray_push_back(xkb->defaultVariants, &dummy);
+
+    if (toDefault) {
+        if (index == 0) {
+            return;
+        }
+        if (index > 0) {
+            utarray_remove_quick_full(xkb->defaultLayouts, index);
+            utarray_remove_quick_full(xkb->defaultVariants, index);
+        }
+        utarray_push_front(xkb->defaultLayouts, &layoutString);
+        if (variantString) {
+            utarray_push_front(xkb->defaultVariants, &variantString);
+        } else {
+            const char *dummy = "";
+            utarray_push_front(xkb->defaultVariants, &dummy);
+        }
+    } else {
+        while (utarray_len(xkb->defaultVariants) >= 4) {
+            utarray_pop_back(xkb->defaultVariants);
+            utarray_pop_back(xkb->defaultLayouts);
+        }
+        utarray_push_back(xkb->defaultLayouts, &layoutString);
+        if (variantString) {
+            utarray_push_back(xkb->defaultVariants, &variantString);
+        } else {
+            const char *dummy = "";
+            utarray_push_back(xkb->defaultVariants, &dummy);
+        }
     }
     FcitxXkbSetLayout(xkb, NULL, NULL, NULL);
 }
 
 static int
-FcitxXkbFindOrAddLayout(FcitxXkb *xkb, const char *layout, const char *variant)
+FcitxXkbFindOrAddLayout(FcitxXkb *xkb, const char *layout, const char *variant, boolean toDefault)
 {
     int index;
     if (layout == NULL)
         return -1;
     index = FcitxXkbFindLayoutIndex(xkb, layout, variant);
-    if (index >= 0)
-        return index;
     if (!xkb->config.bOverrideSystemXKBSettings)
-        return -1;
-    FcitxXkbAddNewLayout(xkb, layout, variant);
+        return index;
+    if (!(index < 0 || (index > 0 && toDefault)))
+        return index;
+    FcitxXkbAddNewLayout(xkb, layout, variant, toDefault, index);
     FcitxXkbInitDefaultLayout(xkb);
     return FcitxXkbFindLayoutIndex(xkb, layout, variant);
 }
 
 static boolean
-FcitxXkbSetLayoutByName(FcitxXkb *xkb, const char *layout, const char *variant)
+FcitxXkbSetLayoutByName(FcitxXkb *xkb, const char *layout, const char *variant, boolean toDefault)
 {
     int index;
-    index = FcitxXkbFindOrAddLayout(xkb, layout, variant);
+    index = FcitxXkbFindOrAddLayout(xkb, layout, variant, toDefault);
     if (index < 0) {
         return false;
     }
@@ -553,10 +618,32 @@ FcitxXkbRetrieveCloseGroup(FcitxXkb *xkb)
 {
     LayoutOverride* item = NULL;
     HASH_FIND_STR(xkb->layoutOverride, "default", item);
-    if (item)
-        FcitxXkbSetLayoutByName(xkb, item->layout, item->variant);
-    else
-        FcitxXkbSetLayoutByName(xkb, xkb->closeLayout, xkb->closeVariant);
+    if (item) {
+        FcitxXkbSetLayoutByName(xkb, item->layout, item->variant, true);
+    } else {
+        do {
+            if (!xkb->config.useFirstKeyboardIMAsDefaultLayout) {
+                break;
+            }
+            char *layout = NULL, *variant = NULL;
+
+            UT_array* imes = FcitxInstanceGetIMEs(xkb->owner);
+            if (utarray_len(imes) == 0) {
+                break;
+            }
+
+            FcitxIM* im = (FcitxIM*) utarray_front(imes);
+            ExtractKeyboardIMLayout(im->uniqueName, &layout, &variant);
+            if (!layout) {
+                break;
+            }
+            FcitxXkbSetLayoutByName(xkb, layout, variant, true);
+            free(layout);
+            free(variant);
+            return;
+        } while(0);
+        FcitxXkbSetLayoutByName(xkb, xkb->closeLayout, xkb->closeVariant, true);
+    }
 }
 
 static void FcitxXkbIMKeyboardLayoutChanged(void* arg, const void* value)
@@ -565,15 +652,10 @@ static void FcitxXkbIMKeyboardLayoutChanged(void* arg, const void* value)
      * if fcitx-xkb doesn't change the layout for it. */
     FcitxXkb *xkb = (FcitxXkb*) arg;
     FcitxIM *currentIM = FcitxInstanceGetCurrentIM(xkb->owner);
-    if (xkb->config.bIgnoreInputMethodLayoutRequest
-        && (!currentIM || strncmp(currentIM->uniqueName, "fcitx-keyboard",
-                                  strlen("fcitx-keyboard")) != 0)) {
-        FcitxXkbRetrieveCloseGroup(xkb);
-        return;
-    }
 
     /* active means im will be take care */
     if (FcitxInstanceGetCurrentStatev2(xkb->owner) == IS_ACTIVE) {
+        char *layoutToFree = NULL, *variantToFree = NULL;
         char* layoutString = NULL;
         char* variantString = NULL;
         LayoutOverride* item = NULL;
@@ -585,24 +667,55 @@ static void FcitxXkbIMKeyboardLayoutChanged(void* arg, const void* value)
             variantString = item->variant;
         }
         else {
-            const char* layout = (const char*) value;
-            if (layout) {
-                s = fcitx_utils_split_string(layout, ',');
-                char  **pLayoutString = (char**)utarray_eltptr(s, 0);
-                char **pVariantString = (char**)utarray_eltptr(s, 1);
-                layoutString = (pLayoutString)? *pLayoutString: NULL;
-                variantString = (pVariantString)? *pVariantString: NULL;
-            }
-            else {
-                layoutString = NULL;
-                variantString = NULL;
+
+            do {
+                if (!xkb->config.useFirstKeyboardIMAsDefaultLayout) {
+                    break;
+                }
+                if (currentIM && IMIsKeyboardIM(currentIM->uniqueName)) {
+                    break;
+                }
+
+                UT_array* imes = FcitxInstanceGetIMEs(xkb->owner);
+                if (utarray_len(imes) == 0) {
+                    break;
+                }
+
+                FcitxIM* im = (FcitxIM*) utarray_front(imes);
+                ExtractKeyboardIMLayout(im->uniqueName, &layoutToFree, &variantToFree);
+                if (!layoutToFree) {
+                    break;
+                }
+
+                layoutString = layoutToFree;
+                variantString = variantToFree;
+
+            } while(0);
+
+            if (!layoutString) {
+                const char* layoutVariantString = (const char*) value;
+                if (layoutVariantString) {
+                    s = fcitx_utils_split_string(layoutVariantString, ',');
+                    char  **pLayoutString = (char**)utarray_eltptr(s, 0);
+                    char **pVariantString = (char**)utarray_eltptr(s, 1);
+                    layoutString = (pLayoutString)? *pLayoutString: NULL;
+                    variantString = (pVariantString)? *pVariantString: NULL;
+                }
+                else {
+                    layoutString = NULL;
+                    variantString = NULL;
+                }
             }
         }
-        if (!FcitxXkbSetLayoutByName(xkb, layoutString, variantString))
+        if (!FcitxXkbSetLayoutByName(xkb, layoutString, variantString, false)) {
             FcitxXkbRetrieveCloseGroup(xkb);
+        }
         if (s) {
             fcitx_utils_free_string_list(s);
         }
+
+        fcitx_utils_free(layoutToFree);
+        fcitx_utils_free(variantToFree);
     } else {
         FcitxXkbRetrieveCloseGroup(xkb);
     }
@@ -642,6 +755,7 @@ FcitxXkbSaveCloseGroup(FcitxXkb *xkb)
     fcitx_utils_free(xkb->closeVariant);
     xkb->closeLayout = tmplayout;
     xkb->closeVariant = tmpvariant;
+    FcitxXkbRetrieveCloseGroup(xkb);
 }
 
 static void
@@ -716,6 +830,13 @@ static void* FcitxXkbCreate(FcitxInstance* instance)
     return NULL;
 }
 
+static void FcitxXkbScheduleRefresh(void* arg) {
+    FcitxXkb* xkb = (FcitxXkb*) arg;
+    FcitxUIUpdateInputWindow(xkb->owner);
+    FcitxXkbInitDefaultLayout(xkb);
+    FcitxXkbApplyCustomScript(xkb);
+}
+
 static boolean FcitxXkbEventHandler(void* arg, XEvent* event)
 {
     FcitxXkb* xkb = (FcitxXkb*) arg;
@@ -725,13 +846,10 @@ static boolean FcitxXkbEventHandler(void* arg, XEvent* event)
 
         if (xkbEvent->any.xkb_type == XkbStateNotify &&
             xkbEvent->state.changed & GROUP_CHANGE_MASK) {
-            FcitxIM* currentIM = FcitxInstanceGetCurrentIM(xkb->owner);
-            if (FcitxInstanceGetCurrentStatev2(xkb->owner) != IS_ACTIVE
-                || (xkb->config.bIgnoreInputMethodLayoutRequest
-                    && (!currentIM ||
-                        strncmp(currentIM->uniqueName, "fcitx-keyboard",
-                                strlen("fcitx-keyboard")) != 0)))
+            if (xkb->config.useFirstKeyboardIMAsDefaultLayout
+                && FcitxInstanceGetCurrentStatev2(xkb->owner) != IS_ACTIVE) {
                 FcitxXkbSaveCloseGroup(xkb);
+            }
         }
 
         if (xkbEvent->any.xkb_type == XkbNewKeyboardNotify
@@ -739,9 +857,8 @@ static boolean FcitxXkbEventHandler(void* arg, XEvent* event)
         ) {
             xkb->lastSerial = xkbEvent->new_kbd.serial;
             XSync(xkb->dpy, False);
-            FcitxUIUpdateInputWindow(xkb->owner);
-            FcitxXkbInitDefaultLayout(xkb);
-            FcitxXkbApplyCustomScript(xkb);
+            FcitxInstanceRemoveTimeoutByFunc(xkb->owner, FcitxXkbScheduleRefresh);
+            FcitxInstanceAddTimeout(xkb->owner, 10, FcitxXkbScheduleRefresh, xkb);
         }
         return true;
     }
@@ -777,6 +894,8 @@ static void FcitxXkbDestroy(void* arg)
     fcitx_utils_free(xkb->closeLayout);
     fcitx_utils_free(xkb->closeVariant);
     fcitx_utils_free(xkb->defaultXmodmapPath);
+
+    FcitxConfigFree(&xkb->config.gconfig);
     free(xkb);
 }
 
@@ -864,8 +983,7 @@ void LoadLayoutOverride(FcitxXkb* xkb)
         LayoutOverride* override = NULL;
         HASH_FIND_STR(xkb->layoutOverride, imString, override);
         /* if rule exists or the rule is for fcitx-keyboard, skip it */
-        if (override || strncmp(imString, "fcitx-keyboard",
-                                strlen("fcitx-keyboard")) == 0) {
+        if (override || IMIsKeyboardIM(imString)) {
             continue;
         }
 
@@ -968,7 +1086,7 @@ FcitxXkbSetLayoutOverride(FcitxXkb *xkb, const char *imname, const char *layout,
     }
 
     if (layout && layout[0] != '\0' &&
-        strncmp(imname, "fcitx-keyboard", strlen("fcitx-keyboard")) != 0) {
+        !IMIsKeyboardIM(imname)) {
         item = fcitx_utils_new(LayoutOverride);
         item->im = strdup(imname);
         item->layout = strdup(layout);
